package main

import (
	"bufio"
	"compress/gzip"
	"flag"
	"fmt"
	"io"
	"io/ioutil"
	"log"
	"os"
	"path"
	"strconv"
	"strings"
)

func check(e error) {
	if e != nil {
		panic(e)
	}
}

// Copy source file to destination.
// Output is automatically gzipped if destination has the ".gz" extension.
// Terminates program if there is an error,
// logs number of bytes copied otherwise.
func copyFile(source string, destination string) {
	sourceFile, err := os.Open(source)
	check(err)
	defer sourceFile.Close()

	os.MkdirAll(path.Dir(destination), 0775)
	destFile, err := os.Create(destination)
	check(err)
	defer destFile.Close()

	gzipOutput := !strings.HasSuffix(source, ".gz") && strings.HasSuffix(destination, ".gz")
	var size int64
	if gzipOutput {
		log.Println("Gzipping", sourceFile.Name())
		gzipWriter := gzip.NewWriter(destFile)
		defer gzipWriter.Close()
		size, err = io.Copy(gzipWriter, sourceFile)
		check(err)
	} else {
		size, err = io.Copy(destFile, sourceFile)
		check(err)
	}
	log.Println("Copied", size, "bytes from", sourceFile.Name(), "to", destFile.Name())
}

// Add U, N, S1, K1, S2, K2, ... at the end of each line in RunInfo.csv. Dunno why.
func convertRunInfoCsv(source string, destination string) {
	sourceFile, err := os.Open(source)
	check(err)
	defer sourceFile.Close()

	os.MkdirAll(path.Dir(destination), 0755)
	destFile, err := os.Create(destination)
	check(err)
	defer destFile.Close()

	scanner := bufio.NewScanner(sourceFile)

	// check and write header
	scanner.Scan()
	if scanner.Text() == "Run,Condition,Onset,Duration" {
		destFile.WriteString("Run,Condition.Onset,Duration,face\n")
	} else {
		log.Fatal("Incorrect format for ", sourceFile.Name())
	}

	i := 1
	repNum := 0
	var outStr string
	var values []string
	foundRep := false

	for scanner.Scan() {
		outStr = scanner.Text()
		values = strings.Split(outStr, ",")
		if values[1] == "rep" {
			if repNum%2 == 0 {
				outStr += ",S" + strconv.Itoa(1+repNum/2)
			} else {
				outStr += ",K" + strconv.Itoa(1+repNum/2)
			}
			repNum++
			foundRep = true
		} else {
			if foundRep {
				outStr += ",N"
			} else {
				outStr += ",U"
			}
			repNum = 0
		}
		destFile.WriteString(outStr + "\n")
		i++
	}
	log.Println("Wrote", i, "lines to", destFile.Name())
}

func convertSubject(inputDir string, outputDir string, gzipFlag bool, onlyAnat bool) {
	if (!onlyAnat) {
		convertRunInfoCsv(inputDir+"/RunInfo.csv", outputDir+"/RunInfo.csv")
	}
	subjectName := path.Base(inputDir)
	suffix := "_reorient.nii"
	if gzipFlag {
		suffix += ".gz"
	}
	copyFile(inputDir+"/MPRAGE/MPRAGE_1.nii",
		outputDir+"/anat/003/"+subjectName+"_mpr003"+suffix)

	if (onlyAnat) {
		return;
	}

	for runId := 1; runId < 7; runId++ {
		bldIdStr := strconv.Itoa(runId + 7)
		if runId < 3 {
			bldIdStr = "00" + bldIdStr
		} else {
			bldIdStr = "0" + bldIdStr
		}
		copyFile(inputDir+"/functional/Orig/run_"+strconv.Itoa(runId)+".nii",
			outputDir+"/bold/"+bldIdStr+"/"+subjectName+"_bld"+bldIdStr+"_rest"+suffix)
	}
}

// Convert all subjects in subdirectories using convertSubject()
func convertAll(inputDir string, outputDir string, gzipFlag bool, templates []string, onlyAnat bool) {
	files, err := ioutil.ReadDir(inputDir)
	check(err)

	for _, f := range files {
		if f.IsDir() {
			inputSubjectDir := inputDir + "/" + f.Name()
			outputSubjectDir := outputDir + "/" + f.Name()
			if _, err := os.Stat(inputSubjectDir + "/" + "RunInfo.csv"); err == nil {
				log.Println("Processing", inputSubjectDir)
				convertSubject(inputSubjectDir, outputSubjectDir, gzipFlag, onlyAnat)
				for i := 0; i < len(templates); i++ {
					copyFile(templates[i], outputSubjectDir+"/"+path.Base(templates[i]))
				}
			} else {
				log.Println("File RunInfo.csv does not exist in " + inputSubjectDir + ". Skipping directory.")
			}
		}
	}
}

func main() {
	// define flags
	gzipFlagPtr := flag.Bool("gzip", false, "Gzip Nifti files in output")
	templatesFlagPtr := flag.String("templates", "",
		"One ore more comma-separated files that will be copied to outputDirectory")
	onlyAnatPtr := flag.Bool("only-anat", false, "Copy only the anatomical data, no functional")
	flag.Parse()

	// check arguments and flags
	arguments := flag.Args()
	if len(arguments) != 2 {
		fmt.Println("Usage: ", path.Base(os.Args[0]), " [-gzip] [-templates=/path/to/file.txt] inputDirectory outputDirectory")
		flag.PrintDefaults()
		os.Exit(0)
	}
	var templates []string
	if len(*templatesFlagPtr) > 0 {
		templates = strings.Split(*templatesFlagPtr, ",")
	}
	log.Println("Number of template files is", len(templates))

	convertAll(path.Clean(arguments[0]), path.Clean(arguments[1]), *gzipFlagPtr, templates, *onlyAnatPtr)
}
